/*MEX implementation */

/*author: Michael Wohlmayr, 19.02.2010 */
/*v.0.1*/

#include "mex.h"
#include "math.h"


void logsumexp(mxArray *X,double *result)
{
    /*for each row in X, compute logsumexp in a numerically stable way.*/
    /*Results are written to vector starting at address specified by 'result'*/
    /*It is assumed that this result vector initially contains zeros!*/
    
    /*VARIABE DECLARATION*/
    mxArray *mexIn[3]; /*For calling MATLAB function*/
    mxArray *mexOut[1]; /*For calling MATLAB function*/
    double *maxVal = 0;
    double *p = 0; /*working pointer*/
    double *p2 = 0; /*working pointer*/
    int nRow = 0;
    int nCol = 0;
    int kRow = 0;
    int kCol = 0;
    
    nRow = mxGetM(X);
    nCol = mxGetN(X);
    
    
    /*Call MATLAB maximum operator function*/
    mexIn[0] = X;
    mexIn[1] = mxCreateDoubleMatrix(0,0,mxREAL); /*second argument is: '[ ]'*/
    mexIn[2] = mxCreateDoubleMatrix(1,1,mxREAL);
    *mxGetPr(mexIn[2]) = 2; /*third argument is 2*/
    
    mexCallMATLAB(1,mexOut,3,mexIn,"max");
    
    /*for each row, sum over exp(x_i - maxVal)*/
    p2 = mxGetPr(X);
    for(kCol=0;kCol<nCol;kCol++) {
        p = result;
        maxVal = mxGetPr(mexOut[0]);
        for(kRow = 0;kRow<nRow; kRow++){
            
            *p += exp(*p2 - *maxVal);
            
            p++;
            maxVal++;
            p2++;
        }
    }
    
    /*take log of the above sum and add maxVal*/
    p = result;
    maxVal = mxGetPr(mexOut[0]);
    for(kRow=0;kRow<nRow;kRow++)
        *p++ = log(*p) + *maxVal++;
        
    
    
    mxDestroyArray(mexIn[1]);
    mxDestroyArray(mexIn[2]);
    mxDestroyArray(mexOut[0]);
    
}


/* The gateway routine */
void mexFunction(int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[]) {
    
    /*VARIABLE DECLARATION*/
    int nClass = 0;
    int kClass = 0;
    int nTrainSample = 0;
    int nTestSample = 0;
    int kSample = 0;
    int nComponent = 0;
    int kComponent = 0;
    int nDim = 0;
    int kDim = 0;
    
    const mxArray *cellArrayMu = 0; /*cell array (1 x nClass) holding nDim x nComponent matrices (nComponent varies from matrix to matrix)*/
    const mxArray *cellArraySigma = 0; /*cell array (1 x nClass) holding nDim x nComponent matrices (nComponent varies from matrix to matrix)*/
    const mxArray *cellArrayW = 0; /*cell array (1 x nClass) holding 1 x nComponent vectors (nComponent varies from matrix to matrix)*/
    const mxArray *trainX = 0; /*matrix (nTrainSample x nDim) holding training samples*/
    const mxArray *testX = 0; /*matrix (nTestSample x nDim) holding test samples*/
    
    mxArray *mu = 0; /*nDim x nComponent matrix*/
    mxArray *sigma = 0; /*nDim x nComponent matrix*/
    mxArray *w = 0; /*nDim x nComponent matrix*/
    
    mxArray *A = 0; /*for intermediate results */
    mxArray *B = 0; /*for intermediate results */
    mxArray *std = 0; /*to precompute square root of covariance matrix*/
    mxArray *logStd = 0; /*to precompute log of square root of covariance matrix*/
    mxArray *logW = 0; /*to precompute log of w*/
    
    mxArray *pXC = 0; /*first result matrix*/
    mxArray *gamma = 0; /*second result matrix*/
    mxArray *testPXC = 0; /*third result matrix*/
    
    double *trainXPtr = 0; /*points to data in trainX*/
    double *testXPtr = 0; /*points to data in testX*/
    double *ptrPXC = 0; /*points to data in pXC*/
    double *ptrTestPXC = 0; /*points to data in testPXC*/
    double *muPtr = 0; /*points to data in mu{k}*/
    double *sigmaPtr = 0; /*points to data in sigma{k}*/
    double *stdPtr = 0; /*points to data in squareroot of sigma{k}*/
    double *logStdPtr = 0; /*points to data in log of squareroot of sigma{k}*/
    double *wPtr = 0; /*points to data in w*/
    double *logWPtr = 0; /*points to data in log of w*/
    double *aPtr = 0; /*points to data in A*/
    double *bPtr = 0; /*points to data in B*/
    double *p = 0; /*pointer for arbitrary purposes*/
    double *p2 = 0; /*pointer for arbitrary purposes*/
 
    double buffer = 0;
    double const1 = 0;
    double pi = 0;
    /*END VARIABLE DECLARATION*/
    
    
    /*ASSIGN, PARSE AND CHECK INPUT PARAMETERS*/
    if(nrhs != 5){
        mexErrMsgTxt("5 arguments expected\n");
        nlhs = 0;
        return;
    }
    
    cellArrayMu = prhs[0];
    cellArraySigma = prhs[1];
    cellArrayW = prhs[2];
    trainX = prhs[3];
    testX = prhs[4];
    
    
    nClass = mxGetN(cellArrayMu);
    if(mxGetN(cellArraySigma) != nClass)
        mexErrMsgTxt("Size of cellArrayMu not equal to cellArraySigma");
    
    nDim = mxGetM(mxGetCell(cellArrayMu, 0)); /*Get dimension nDim*/
    nTrainSample = mxGetM(trainX); /*Get number of training samples*/
    nTestSample = mxGetM(testX); /*Get number of test samples*/
    
    
    pXC = mxCreateDoubleMatrix(nTrainSample, nClass, mxREAL); /*Result matrix for p(X|C) on training data*/
    ptrPXC = mxGetPr(pXC);
    
    testPXC = mxCreateDoubleMatrix(nTestSample, nClass, mxREAL); /*Result matrix for p(X|C) on test data*/
    ptrTestPXC = mxGetPr(testPXC);
    
    
    gamma = mxCreateCellMatrix(1,nClass); /*set cells with mxSetCell */
    
    pi = 4*atan(1);
    const1 = nDim*0.5*log(2*pi);
    
    /*LOOP OVER ALL CLASSES*/
    for(kClass=0;kClass<nClass;kClass++) {
              
        /*get GMM parameters for that class */
        mu = mxGetCell(cellArrayMu, kClass);
        sigma = mxGetCell(cellArraySigma, kClass);
        w = mxGetCell(cellArrayW, kClass);
        muPtr = mxGetPr(mu);
        sigmaPtr = mxGetPr(sigma);
        wPtr = mxGetPr(w);
        nComponent = mxGetN(mu);
        
        /*Precompute standard deviation and log of it*/
        /*Precompute log w*/
        std = mxCreateDoubleMatrix(nDim,nComponent,mxREAL);
        stdPtr = mxGetPr(std);
        logStd = mxCreateDoubleMatrix(nDim,nComponent,mxREAL);
        logStdPtr = mxGetPr(logStd);
        logW = mxCreateDoubleMatrix(1,nComponent,mxREAL);
        logWPtr = mxGetPr(logW);
        for(kComponent=0;kComponent<nComponent;kComponent++) {
            
            *logWPtr = log(*wPtr);
            logWPtr++;
            wPtr++;
            
            for(kDim=0;kDim<nDim;kDim++) {
                *stdPtr = sqrt(*sigmaPtr);
                *logStdPtr = 0.5*log(*sigmaPtr);
                
                sigmaPtr++;
                stdPtr++;
                logStdPtr++;   
            }         
        }
        sigmaPtr = mxGetPr(sigma);
        stdPtr = mxGetPr(std);
        logStdPtr = mxGetPr(logStd);
        logWPtr = mxGetPr(logW);

        /* create workspace for intermediate results*/
        A = mxCreateDoubleMatrix(nTrainSample,nComponent,mxREAL); /*will hold gamma for actual class*/
        aPtr = mxGetPr(A);
        B = mxCreateDoubleMatrix(nTestSample,nComponent,mxREAL); /*will hold intermediate results for testData*/
        bPtr = mxGetPr(B);
        
        /*start computations*/
        for(kComponent=0;kComponent<nComponent;kComponent++) {
            trainXPtr = mxGetPr(trainX);    
            testXPtr = mxGetPr(testX);   
            
            /*for this component, compute the log normal distribution of training and test data by summing over all dimensions. Results are stored in A and B, respectively.*/
            for(kDim=0;kDim<nDim;kDim++) {
             
                p = aPtr; /*points to A(kSample,kComponent) (MATLAB style)*/
                for(kSample=0;kSample<nTrainSample;kSample++) {
                    /*training data*/
                    buffer = ((*trainXPtr)-(*muPtr))/(*stdPtr); /*standardize sample*/
                    buffer = 0.5*buffer*buffer + (*logStdPtr);
                    *p += buffer; 
                    
                    p++; 
                    trainXPtr++; /*next sample of that dimension*/  
                }
                
                p = bPtr;
                for(kSample=0;kSample<nTestSample;kSample++) {
                    /*test data*/
                    buffer = ((*testXPtr)-(*muPtr))/(*stdPtr); /*standardize sample*/
                    buffer = 0.5*buffer*buffer + (*logStdPtr);
                    *p += buffer; 
                    
                    p++; 
                    testXPtr++; /*next sample of that dimension*/  
                }
                
                muPtr++; /*move to next position in mu matrix.*/
                stdPtr++;
                logStdPtr++;
            }
            
            
            /*Subtract const1 and add log w. As a result, we have log p(x,k|c) for each sample x and each component k.*/
            p = aPtr;
            for(kSample=0;kSample<nTrainSample;kSample++) {
                buffer = *p;
                buffer = -buffer-const1;
                buffer += *logWPtr;
                *p = buffer;
                p++;
            }
            
            p = bPtr;
            for(kSample=0;kSample<nTestSample;kSample++) {
                buffer = *p;
                buffer = -buffer-const1;
                buffer += *logWPtr;
                *p = buffer;           
                p++;
            }
            
            aPtr += nTrainSample; /*now points to A(1,kComponent+1) (MATLAB style)*/ 
            bPtr += nTestSample;
            logWPtr++; /*now points to logW(kComponent+1) (MATLAB style)*/
        } /*end for loop over kComponent*/
        
        
        logsumexp(A,ptrPXC); /*FIXXME from here on*/
        logsumexp(B,ptrTestPXC);
        
        /*Now we normalize A and apply exp(.) to obtain gamma. The normalization is log p(x|c), which is already stored in pXC.*/
        aPtr = mxGetPr(A);
        for(kComponent=0;kComponent<nComponent;kComponent++) {
            p = ptrPXC;
            for(kSample=0;kSample<nTrainSample;kSample++) {
                *aPtr++ = exp(*aPtr - *p++);
            }
        }

        ptrPXC += nTrainSample; /*set to first element of (kClass+1)-th column*/
        ptrTestPXC += nTestSample; /*set to first element of (kClass+1)-th column*/

        mxSetCell(gamma,kClass,A); /*assign A to particular cell in cell array*/
        
        /*dealloc memory for intermediate results*/
        mxDestroyArray(std);
        std = 0;
        mxDestroyArray(logStd);
        logStd = 0;
        mxDestroyArray(logW);
        logW = 0;
        mxDestroyArray(B);
        B = 0;
        /*Do NOT DESTROY array A, as we passed it's pointer over to cell array gamma.*/
        
    }
    
    plhs[0] = pXC;
    plhs[1] = gamma;
    plhs[2] = testPXC;
    
    return;
}
